"""Implements multinomial and mixed logit models."""

import logging
import warnings
from abc import ABC, abstractmethod
from time import time

import numpy as np
import scipy as sc
from scipy.stats import norm, t

from ._device import device as dev
from .boxcox_functions import boxcox_param_deriv, boxcox_transformation

logger = logging.getLogger(__name__)

warnings.simplefilter(action='ignore', category=FutureWarning)
"""
Notations
---------
    N : Number of choice situations
    P : Number of observations per panels
    J : Number of alternatives
    K : Number of variables (Kf: fixed + non-trans, Kr: random+non-trans,
                            Kftrans: fixed+trans, Krtrans: random+trans)
"""


class ChoiceModel(ABC):
    """Base class for estimation of discrete choice models"""

    def __init__(self):
        """Init Function

        Parameters
        ----------
        random_state: an integer used as seed to generate numpy random numbers
        """
        self.coeff_names = None
        self.coeff_ = None
        self.stderr = None
        self.zvalues = None
        self.pvalues = None
        self.loglikelihood = None
        self.total_fun_eval = 0
        self.verbose = 1

    def _reset_attributes(self):
        self.coeff_names = None
        self.coeff_ = None
        self.stderr = None
        self.zvalues = None
        self.pvalues = None
        self.loglikelihood = None
        self.total_fun_eval = 0
        self.verbose = 1

    @abstractmethod
    def fit(self, X, y, varnames=None, alts=None, isvars=None, transvars=None,
            transformation=None, ids=None,
            weights=None, base_alt=None, fit_intercept=False, init_coeff=None,
            maxiter=2000, random_state=None, correlation=None):
        pass

    def _as_array(self, X, y, varnames, alts, isvars, transvars, ids,
                  weights, panels, avail):
        X = np.asarray(X)
        y = np.asarray(y)
        varnames = np.asarray(varnames) if varnames is not None else None
        alts = np.asarray(alts) if alts is not None else None
        isvars = np.asarray(isvars) if isvars is not None else None
        transvars = np.asarray(transvars) if transvars is not None else []
        ids = np.asarray(ids) if ids is not None else None
        weights = np.asarray(weights) if weights is not None else None
        panels = np.asarray(panels) if panels is not None else None
        avail = np.asarray(avail) if avail is not None else None
        return X, y, varnames, alts, isvars, transvars, ids, \
            weights, panels, avail

    def _pre_fit(self, alts, varnames, isvars, transvars, base_alt,
                 fit_intercept, transformation, maxiter, panels=None,
                 correlation=None, randvars=None):
        self._reset_attributes()
        self._fit_start_time = time()
        self.isvars = [] if isvars is None else isvars
        self.transvars = [] if transvars is None else transvars
        self.randvars = [] if randvars is None else randvars
        with warnings.catch_warnings():
            # CURRENTLY IGNORING FUTURE WARNING
            # CURRENT PY: 3.8.3, numpy: 1.18.5
            warnings.simplefilter(action='ignore', category=FutureWarning)
            self.asvars = [v for v in varnames if ((v not in self.isvars) and
                           (v not in self.transvars) and
                           (v not in self.randvars))]

            self.asvars_construct_matrix = [v for v in varnames  # old definition of asvars used to make datasets
                                            if ((v not in self.isvars))]
        self.randtransvars = [] if transvars is None else []
        self.fixedtransvars = [] if transvars is None else []
        self.alternatives = np.unique(alts)

        self.varnames = list(varnames)  # Easier to handle with lists
        self.fit_intercept = fit_intercept
        self.transformation = transformation
        self.base_alt = self.alternatives[0] if base_alt is None else base_alt
        self.correlation = False if correlation is None else correlation
        self.maxiter = maxiter
        if not hasattr(self, 'panels'):  # check panels not already set (LCM)
            self.panels = panels

    def _post_fit(self, optimization_res, coeff_names, sample_size, hess_inv=None, verbose=1):
        self.convergence = optimization_res['success']
        self.coeff_ = optimization_res['x']
        # convert hess inverse for L-BFGS-B optimisation method

        self.stderr = np.zeros_like(self.coeff_)
        std_err_estimated = False

        if 'stderr' in optimization_res:
            std_err_estimated = True

            self.stderr = optimization_res['stderr']

        is_latent_class = optimization_res['is_latent_class'] if \
            'is_latent_class' in optimization_res else False
        self.is_latent_class = is_latent_class
        if is_latent_class:
            new_coeff_names = np.array([])
            for i in range(self.num_classes):
                X_class_idx = self._get_class_X_idx(i, coeff_names=coeff_names)
                class_coeff_names = coeff_names[X_class_idx]
                class_coeff_names = np.core.defchararray.add('class-' + str(i+1) +
                                                            ': ', class_coeff_names)
                new_coeff_names = np.concatenate((new_coeff_names, class_coeff_names))
            coeff_names = new_coeff_names

        if hasattr(self, 'Hinv') and not is_latent_class:
            if dev.using_gpu:
                self.stderr = np.sqrt(np.diag(self.Hinv))
            else:
                diag_arr_tmp = np.diag(np.array(self.Hinv))
                # stop runtimewarnings from (very small) negative values
                # assume these occur from some floating point error and are 0.
                pos_vals_idx = [ii for ii, el in enumerate(diag_arr_tmp)
                                if el > 0]
                diag_arr = np.zeros(len(diag_arr_tmp))
                diag_arr[pos_vals_idx] = diag_arr_tmp[pos_vals_idx]
                self.stderr = np.sqrt(diag_arr)
            std_err_estimated = False if np.isnan(self.stderr).any else True

        if not std_err_estimated:
            if self.method == "bfgs":
                self.stderr = np.sqrt(np.diag(optimization_res['hess_inv']))
            if self.method == "l-bfgs-b":
                hess = optimization_res['hess_inv'].todense()
                self.stderr = np.sqrt(np.diag(np.array(hess)))

        lambda_mask = [1 if "lambda" in x else 0 for x in coeff_names]

        if len(lambda_mask) != len(self.coeff_):
            lambda_mask = np.ones_like(self.coeff_)

        if 'is_latent_class' in optimization_res:
            lambda_mask = np.zeros_like(self.coeff_)

        self.num_params = self.Kbw + self.Kchol + self.Kf + self.Kftrans + self.Kr + self.Krtrans
        self.zvalues = np.nan_to_num((self.coeff_ - lambda_mask)/self.stderr)
        # set maximum (and minimum) limits to zvalues
        self.zvalues = [z if z < 1e+5 else 1e+5 for z in self.zvalues]
        self.zvalues = [z if z > -1e+5 else -1e+5 for z in self.zvalues]
        if sample_size < 100:  # arbitrary ... could do standard 30
            self.pvalues = 2*(1 - t.cdf(np.abs(self.zvalues), df=sample_size))
        else:
            self.pvalues = 2*(1 - norm.cdf(np.abs(self.zvalues)))
        self.loglikelihood = -optimization_res['fun']
        self.coeff_names = coeff_names
        self.total_iter = optimization_res['nit']
        self.estim_time_sec = time() - self._fit_start_time
        self.sample_size = sample_size
        num_params = len(self.coeff_)
        if self.is_latent_class:
            num_params = len(self.coeff_) + len(optimization_res['class_x'])
        self.aic = 2*num_params - 2*self.loglikelihood
        self.bic = np.log(sample_size)*num_params - 2*self.loglikelihood

        if 'is_latent_class' in optimization_res:
            self.is_latent_class = True
            self.class_x = optimization_res['class_x']
            self.class_x_stderr = optimization_res['class_x_stderr']

        if not self.convergence and verbose > 0:
            logger.warning("**** The optimization did not converge after {} "
                           "iterations. ****".format(self.total_iter))
            if hasattr(optimization_res, 'message'):
                logger.info("Message: " + str(optimization_res['message']))

    def _setup_design_matrix(self, X):
        """Setups and reshapes input data after adding isvars and intercept.

        Setup the design matrix by adding the intercept when necessary and
        converting the isvars to a dummy representation that removes the base
        alternative.
        """
        if not hasattr(self, 'J'):
            J = len(self.alternatives)
        else:
            J = self.J
        N = P_N = int(len(X)/J)
        self.P = 0
        self.N = N
        self.J = J
        if self.panels is not None:
            # panels size.
            self.P_i = ((np.unique(self.panels, return_counts=True)[1])/J).astype(int)
            self.P = np.max(self.P_i)
            self.N = len(self.P_i)
        else:
            self.P = 1
            self.P_i = np.ones([N]).astype(int)
        isvars = self.isvars.copy()
        asvars = self.asvars.copy()
        asvars_construct_matrix = self.asvars_construct_matrix.copy()
        randvars = self.randvars.copy()
        randtransvars = self.randtransvars.copy()
        fixedtransvars = self.fixedtransvars.copy()
        varnames = self.varnames.copy()
        self.varnames = np.array(varnames)
        ispos = [self.varnames.tolist().index(i) for i in self.isvars]  # Position of IS vars

        # adjust index array to include isvars
        if len(self.isvars) > 0 and not hasattr(self, 'ispos'):  # check not done before...
            self.fxidx = np.insert(np.array(self.fxidx, dtype="bool_"), 0,
                                   np.repeat(True, len(self.isvars)*(J - 1)))
            self.fxtransidx = np.insert(np.array(self.fxtransidx, dtype="bool_"),
                                        0, np.repeat(False, len(self.isvars)*(J - 1)))
            if hasattr(self, 'rvidx'):
                self.rvidx = np.insert(np.array(self.rvidx, dtype="bool_"), 0,
                                       np.repeat(False, len(self.isvars)*(J - 1)))
            if hasattr(self, 'rvtransidx'):
                self.rvtransidx = np.insert(np.array(self.rvtransidx, dtype="bool_"),
                                            0, np.repeat(False, len(self.isvars)*(J - 1)))

        if self.fit_intercept:
            X = np.hstack((np.ones(J*N)[:, None], X))
            if '_inter' not in self.isvars:  # stop running in validation
                # adjust variables to allow intercept parameters
                self.isvars = np.insert(np.array(self.isvars, dtype="<U64"), 0, '_inter')
                self.varnames = np.insert(np.array(self.varnames, dtype="<U64"), 0, '_inter')
                self.fxidx = np.insert(np.array(self.fxidx, dtype="bool_"), 0, np.repeat(True, J-1))
                if hasattr(self, 'rvidx'):
                    self.rvidx = np.insert(np.array(self.rvidx, dtype="bool_"), 0, np.repeat(False, J-1))
                self.fxtransidx = np.insert(np.array(self.fxtransidx, dtype="bool_"), 0, np.repeat(False, J-1))
                if hasattr(self, 'rvtransidx'):
                    self.rvtransidx = np.insert(np.array(self.rvtransidx, dtype="bool_"), 0, np.repeat(False, J-1))

        if self.transformation == "boxcox":
            self.trans_func = boxcox_transformation
            self.transform_deriv = boxcox_param_deriv

        S = np.zeros((self.N, self.P, self.J))
        for i in range(self.N):
            S[i, 0:self.P_i[i], :] = 1
        self.S = S
        ispos = [self.varnames.tolist().index(i) for i in self.isvars]  # Position of IS vars
        aspos = [self.varnames.tolist().index(i) for i in asvars_construct_matrix]  # Position of AS vars
        self.aspos = np.array(aspos) # saved for later use
        self.ispos = np.array(ispos)
        randpos = [self.varnames.tolist().index(i) for i in randvars]  # Position of AS vars
        randtranspos = [self.varnames.tolist().index(i) for i in randtransvars]  # bc transformed variables with random coeffs
        fixedtranspos = [self.varnames.tolist().index(i) for i in fixedtransvars]  # bc transformed variables with fixed coeffs

        self.correlationpos = []
        if randvars:
            self.correlationpos = [self.varnames.tolist().index(x) for x in self.varnames if x in self.randvars]  #  Position of correlated variables within randvars
        if (isinstance(self.correlation, list)):
            self.correlationpos = [self.varnames.tolist().index(x) for x in
                                   self.varnames if x in self.correlation]
            self.uncorrelatedpos = [self.varnames.tolist().index(x) for x in
                                   self.varnames if x not in self.correlation]

        self.Kf = sum(self.fxidx)  # set number of fixed coeffs from idx
        self.Kr = len(randpos)  # Number of random coefficients
        self.Kftrans = len(fixedtranspos)  # Number of fixed coefficients of bc transformed vars
        self.Krtrans = len(randtranspos)  # Number of random coefficients of bc transformed vars
        self.Kchol = 0  # Number of random beta cholesky factors
        self.correlationLength = 0
        self.Kbw = self.Kr

        # set up length of betas required to estimate correlation and/or
        # random variable standard deviations, useful for cholesky matrix
        if (self.correlation):
            if (isinstance(self.correlation, list)):
                self.correlationLength = len(self.correlation)
                self.Kbw = self.Kr - len(self.correlation)
            else:
                self.correlationLength = self.Kr
                self.Kbw = 0
        if (self.correlation):
            if (isinstance(self.correlation, list)):
                # Kchol, permutations of specified params in correlation list
                self.Kchol = int((len(self.correlation) *
                                 (len(self.correlation)+1))/2)
            else:
                # i.e. correlation = True, Kchol permutations of rand vars
                self.Kchol = int((len(self.randvars) *
                                 (len(self.randvars)+1))/2)
        # Create design matrix
        # For individual specific variables
        Xis = None
        if len(self.isvars):
            # Create a dummy individual specific variables for the alts
            dummy = np.tile(np.eye(J), reps=(P_N, 1))
            # Remove base alternative
            dummy = np.delete(dummy,
                              np.where(self.alternatives == self.base_alt)[0],
                              axis=1)
            Xis = X[:, ispos]
            # Multiply dummy representation by the individual specific data
            Xis = np.einsum('nj,nk->njk', Xis, dummy, dtype="float64")
            Xis = Xis.reshape(P_N, self.J, (self.J-1)*len(ispos))
        else:
            Xis = np.array([])
        # For alternative specific variables
        Xas = None
        if asvars_construct_matrix:
            Xas = X[:, aspos]
            Xas = Xas.reshape(N, J, -1)

        # Set design matrix based on existance of asvars and isvars
        if len(asvars_construct_matrix) and len(self.isvars):
            X = np.dstack((Xis, Xas))
        elif len(asvars_construct_matrix):
            X = Xas
        elif (len(self.isvars)):
            X = Xis
        else:
            x_varname_length = len(self.varnames) if not self.fit_intercept \
                               else (len(self.varnames) - 1)+(J-1)
            X = X.reshape(-1, len(self.alternatives), x_varname_length)

        intercept_names = ["_intercept.{}".format(j) for j in self.alternatives
                           if j != self.base_alt] if self.fit_intercept else []
        names = ["{}.{}".format(isvar, j) for isvar in isvars
                 for j in self.alternatives if j != self.base_alt]
        lambda_names_fixed = ["lambda.{}".format(transvar) for transvar in
                              fixedtransvars]
        lambda_names_rand = ["lambda.{}".format(transvar) for transvar in
                             randtransvars]
        randvars = [x for x in self.varnames if x in self.randvars]
        asvars_names = [x for x in asvars if (x not in self.randvars) and
                                             (x not in fixedtransvars) and
                                             (x not in randtransvars)]
        chol = ["chol." + self.varnames[self.correlationpos[i]] + "." +
                self.varnames[self.correlationpos[j]] for i
                in range(self.correlationLength) for j in range(i+1)]
        br_w_names = []
        # three cases for corr. varnames: no corr, corr list, corr Bool (All)
        if (self.correlation is not True and not isinstance(self.correlation, list)):
            if(hasattr(self, "rvidx")):  # avoid errors with multinomial logit
                br_w_names = np.char.add("sd.", randvars)
        if (isinstance(self.correlation, list)):  # if not all r.v.s correlated
            sd_uncorrelated_pos = [self.varnames.tolist().index(x)
                                   for x in self.varnames
                                   if x not in self.correlation and
                                   x in randvars]
            br_w_names = np.char.add("sd.", self.varnames[sd_uncorrelated_pos])
        sd_rand_trans = np.char.add("sd.", self.varnames[randtranspos])
        names = np.concatenate((intercept_names, names, asvars_names, randvars,
                                chol, br_w_names, fixedtransvars,
                                lambda_names_fixed, randtransvars,
                                sd_rand_trans, lambda_names_rand))
        names = np.array(names)
        return X, names

    def _check_long_format_consistency(self, ids, alts, sorted_idx):
        alts = alts[sorted_idx]
        uq_alt = np.unique(alts)
        expect_alt = np.tile(uq_alt, int(len(ids)/len(uq_alt)))
        if not np.array_equal(alts, expect_alt):
            raise ValueError('inconsistent alts values in long format')
        _, obs_by_id = np.unique(ids, return_counts=True)
        if not np.all(obs_by_id/len(uq_alt)):  # Multiple of J
            raise ValueError('inconsistent alts and ids values in long format')

    def _arrange_long_format(self, X, y, ids, alts, panels=None):
        if ids is not None:
            pnl = panels if panels is not None else np.ones(len(ids))
            alts = alts.astype(str)
            alts = alts if len(alts) == len(ids)\
                else np.tile(alts, int(len(ids)/len(alts)))
            cols = np.zeros(len(ids), dtype={'names': ['panels', 'ids', 'alts'],
                                             'formats': ['<f4', '<f4', '<U64']})
            cols['panels'], cols['ids'], cols['alts'] = pnl, ids, alts
            sorted_idx = np.argsort(cols, order=['panels', 'ids', 'alts'])
            X, y = X[sorted_idx], y[sorted_idx]
            if panels is not None:
                panels = panels[sorted_idx]
            self._check_long_format_consistency(ids, alts, sorted_idx)
        return X, y, panels

    def _validate_inputs(self, X, y, alts, varnames, isvars, ids, weights,
                         panels, base_alt, fit_intercept, maxiter):
        if varnames is None:
            raise ValueError('The parameter varnames is required')
        if alts is None:
            raise ValueError('The parameter alternatives is required')
        if X.ndim != 2:
            raise ValueError("X must be an array of two dimensions in "
                             "long format")
        if y.ndim != 1:
            raise ValueError("y must be an array of one dimension in "
                             "long format")
        if len(varnames) != X.shape[1]:
            raise ValueError("The length of varnames must match the number "
                             "of columns in X")

    def summary(self):
        """
        Prints in console the coefficients and additional estimation outputs
        """
        if self.coeff_ is None:
            warnings.warn("The current model has not been yet estimated",
                          UserWarning)
            return
        if not self.convergence:
            print("-"*50)
            print("WARNING: Convergence was not reached during estimation. "
                  "The given estimates may not be reliable")
            if hasattr(self, "gtol_res"):
                print("gtol:", self.gtol)
                print("Final gradient norm:", self.gtol_res)
            print('*'*50)
        print("Estimation time= {:.1f} seconds".format(self.estim_time_sec))

        if hasattr(self, 'pred_prob'):
            print("Proportion of alternatives: observed choice")
            print(self.obs_prob)
            print("Proportion of alternatives: predicted choice")
            print(self.pred_prob)

        if hasattr(self, 'class_freq'):
            print("Estimated proportion of classes")
            print(self.class_freq)

        fmt = "{:19} {:13.10f} {:13.10f} {:13.10f} {:13.3g} {:3}"
        coeff_name_str_length = 19
        if self.is_latent_class:
            coeff_name_str_length = 28
            print("-"*84)
            fmt = "{:28} {:13.10f} {:13.10f} {:13.10f} {:13.3g} {:3}"
            print("{:28} {:>13} {:>13} {:>13} {:>13}"
              .format("Coefficient", "Estimate", "Std.Err.", "z-val", "P>|z|"))
            print("-"*84)
        else:
            print("-"*75)
            print("{:19} {:>13} {:>13} {:>13} {:>13}"
              .format("Coefficient", "Estimate", "Std.Err.", "z-val", "P>|z|"))
            print("-"*75)
        for i in range(len(self.coeff_)):
            signif = ""
            if self.pvalues[i] < 0.001:
                signif = "***"
            elif self.pvalues[i] < 0.01:
                signif = "**"
            elif self.pvalues[i] < 0.05:
                signif = "*"
            elif self.pvalues[i] < 0.1:
                signif = "."
            print(fmt.format(self.coeff_names[i][:coeff_name_str_length],
                             self.coeff_[i], self.stderr[i], self.zvalues[i],
                             self.pvalues[i], signif))
        if self.is_latent_class:
            zvalues = np.nan_to_num(self.class_x/self.class_x_stderr)
            # set maximum (and minimum) limits to zvalues
            zvalues = [z if z > -1e+5 else -1e+5 for z in zvalues]
            pvalues = 2*(1 - t.cdf(np.abs(zvalues), df=self.sample_size))
            self.pvalues_member = pvalues
            coeff_names_member = np.array([])
            for ii, member_class in enumerate(self.member_params_spec):
                # logic for isvars
                # Remove lambda coeffs from member class param names
                member_class_names_idx = self._get_member_X_idx(ii, coeff_names=self.Xnames)
                lambda_idx = np.where(np.char.find(self.Xnames, 'lambda') != -1)[0]
                sd_idx = np.where(np.char.find(self.Xnames, 'sd') != -1)[0]
                chol_idx = np.where(np.char.find(self.Xnames, 'chol') != -1)[0]
                member_class_names_idx = [x for x in member_class_names_idx
                            if x not in sd_idx and x not in chol_idx
                            and x not in lambda_idx]
                member_class_names_idx = np.sort(member_class_names_idx)
                member_class_names_idx = np.array(member_class_names_idx, dtype='int32')
                member_class_names = self.Xnames[member_class_names_idx]
                if self.membership_as_probability:
                    member_class_names = ["probability"]

                class_coeff_names = np.core.defchararray.add('class-' +
                                                             str(ii+2) +
                                                             ': ',
                                                             member_class_names
                                                             )
                if '_inter' in self.member_params_spec[ii]:
                    inter_name = 'class-' + str(ii+2) + ': ' + 'constant'
                    class_coeff_names = np.concatenate(([inter_name], class_coeff_names))
                coeff_names_member = np.concatenate((coeff_names_member, class_coeff_names))
            self.coeff_names_member = coeff_names_member
            print("-"*84)
            print("{:30} {:>13} {:>13} {:>13} {:>13}"
                  .format("Class Member Coeff", "Estimate", "Std.Err.", "z-val", "P>|z|"))
            print("-"*84)
            for ii, coeff_name in enumerate(coeff_names_member):
                signif = ""
                if pvalues[ii] < 0.001:
                    signif = "***"
                elif pvalues[ii] < 0.01:
                    signif = "**"
                elif pvalues[ii] < 0.05:
                    signif = "*"
                elif pvalues[ii] < 0.1:
                    signif = "."
                # note below: offset coeffnames by num_params to ignore class0
                print(fmt.format(coeff_name[:30], self.class_x[ii],
                                 self.class_x_stderr[ii], zvalues[ii],
                                 pvalues[ii], signif
                                 ))

        print("-"*84) if self.is_latent_class else print("-"*75)
        print("Significance:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1")
        print("")
        print("Log-Likelihood= {:.3f}".format(self.loglikelihood))
        print("AIC= {:.3f}".format(self.aic))
        print("BIC= {:.3f}".format(self.bic))
        loglik_null = self._compute_null_loglik()
        adjust_lik_ratio = 1 - (self.aic / loglik_null)
        self.adjust_lik_ratio = adjust_lik_ratio
        print("Adjusted likelihood ratio index: {:.3f}".format(adjust_lik_ratio))

    def corr(self):
        corr_varnames = [self.varnames[pos] for pos in self.correlationpos]
        K = len(corr_varnames) + 1  # + 1 for coeff names row/col
        str_mat = np.array([])
        # top row of coef names
        str_mat = np.append(str_mat, np.array([''] + corr_varnames))
        self.corr_mat = np.round(self.corr_mat[0:len(corr_varnames),
                                               0:len(corr_varnames)], 8)
        fmt = "{:11}"
        print("correlation matrix")
        if dev.using_gpu:
            self.corr_mat = dev.to_cpu(self.corr_mat)
        for ii, row in enumerate(self.corr_mat):
            str_mat = np.append(str_mat, corr_varnames[ii])
            str_mat = np.append(str_mat, np.array(row))
        str_mat = str_mat.reshape((K, K))
        for row in str_mat:
            for el in row:
                print(fmt.format(el), end='  ')
            print('')

    def cov(self):
        corr_varnames = [self.varnames[pos] for pos in self.correlationpos]
        K = len(corr_varnames) + 1 # + 1 for coeff names row/col
        str_mat = np.array([])
        # top row of coef names
        str_mat = np.append(str_mat, np.array([''] + corr_varnames))
        #
        cov_mat = np.round(self.omega[0:len(corr_varnames), 0:len(corr_varnames)], 8)
        fmt = "{:11}"
        print("covariance matrix")
        if dev.using_gpu:
            cov_mat = dev.to_cpu(cov_mat)
        for ii, row in enumerate(cov_mat):
            str_mat = np.append(str_mat, corr_varnames[ii])
            str_mat = np.append(str_mat, np.array(row))
        str_mat = str_mat.reshape((K, K))
        for row in str_mat:
            for el in row:
                print(fmt.format(el), end='  ')
            print('')

    def fitted(self, type="parameters"):

        # for k in range(self.Kr):
        #     s_jn = np.mean(self.pch2_sd_test[:, k])
        #     s_jn2 = np.sum(self.pch2_sd_test[:, k])

        if type == "parameters":
            if hasattr(self, 'pch2_res'):
                return self.pch2_res
        return

    def stddev(self):
        fmt = "{:11}"

        diags = np.round(np.sqrt(np.diag(self.omega)), 8)
        corr_varnames = [self.varnames[pos] for pos in self.correlationpos]
        rv_names_noncorr = [name for name in self.varnames
                            if name in self.randvars and name not in
                            corr_varnames]
        rvtrans_names = [name for name in self.varnames
                         if name in self.randtransvars]

        distributions_corr = [self.randvarsdict[name] for name in
                              corr_varnames]
        distributions_rv = [self.randvarsdict[name] for name in rv_names_noncorr]
        distributions_rvtrans = [self.randvarsdict[name] for name in
                                 rvtrans_names]
        distributions = distributions_corr + distributions_rv + distributions_rvtrans
        print('Standard Deviations')
        stdevs = np.zeros(len(diags))

        means = self.betas[self.Kf:self.Kf+self.Kr]
        for ii, val in enumerate(diags):
            dist = distributions[ii]
            if dist == 'n':
                stdev = val
            if dist == 'ln':
                stdev = np.sqrt(np.exp(val ** 2) - 1) * \
                    np.exp(means[ii] + 0.5 * val ** 2)
            if dist == 'u':
                stdev = (val ** 2)/3
            if dist == 't':
                stdev = val
            stdevs[ii] = np.round(stdev, 8)

        rv_names_all = corr_varnames + rv_names_noncorr + rvtrans_names
        for name in rv_names_all:
            print(fmt.format(name), end='  ')
        print('')
        for std in stdevs:
            print(fmt.format(std), end='  ')
        print('')
        pass



import numpy

_gpu_available = False
try:
    import cupy
    _gpu_available = True
except ImportError:
    pass


class Device():
    def __init__(self):
        self.np = numpy
        self._using_gpu = False
        if _gpu_available:
            self.np = cupy
            self._using_gpu = True

    def cust_einsum(self, expr, a, b):
        """Efficient einsum for common expressions"""
        # TODO? batch across samples from xlogit
        if True:
            if expr == 'npjk,nkr -> npjr':
                n, p, j, k = a.shape
                r = b.shape[-1]
                return self.np.matmul(a.reshape(n, p*j, k), b).reshape(n, p, j, r)
            elif expr == 'npjk,k -> npj':
                return self.np.matmul(a, b)
            elif expr == 'npjr,npjk -> nkr':
                n, p, j, r = a.shape
                k = b.shape[-1]
                return self.np.matmul(
                    b.reshape(n, p*j, k).transpose([0, 2, 1]),
                    a.reshape(n, p*j, r))
            else:
                raise Exception(f"The expression {expr} is not supported by "
                                 "custeinsum")
        return self.np.einsum(expr, a, b)

    def enable_gpu_acceleration(self, device_id=0):
        if(_gpu_available):
            self.np = cupy
            self._using_gpu = True
        else:
            raise Exception("CuPy not found. Verify it is properly installed")
        cupy.cuda.Device(device_id).use()

    def disable_gpu_acceleration(self):
        self.np = numpy
        self._using_gpu = False

    @property
    def using_gpu(self):
        return self._using_gpu

    def to_cpu(self, arr):
        return cupy.asnumpy(arr)

    def to_gpu(self, arr):
        return cupy.asarray(arr)

    def get_device_count(self):
        if _gpu_available:
            return cupy.cuda.runtime.getDeviceCount()
        else:
            return 0


device = Device()



"""Implements all the logic for mixed logit models."""
import itertools
import logging

import numpy as np
import scipy.stats as ss
from scipy.optimize import minimize

from ._choice_model import ChoiceModel
from ._device import device as dev
from .boxcox_functions import (boxcox_param_deriv_mixed,
                               boxcox_transformation_mixed)
from .multinomial_logit import MultinomialLogit

"""
Notations
---------
    N : Number of choice situations
    P : Number of observations per panel
    J : Number of alternatives
    K : Number of variables (Kf: fixed (non-trans), Kr: random,
                             Kftrans: fixed trans, Krtrans: random trans)
"""

# define the computation boundary values not to be exceeded
max_exp_val = 700

max_comp_val = 1e+300
min_comp_val = 1e-300

logger = logging.getLogger(__name__)


class MixedLogit(ChoiceModel):
    """Class for estimation of Mixed Logit Models

    Attributes
    ----------
        coeff_ : numpy array, shape (n_variables + n_randvars, )
            Estimated coefficients

        coeff_names : numpy array, shape (n_variables + n_randvars, )
            Names of the estimated coefficients

        stderr : numpy array, shape (n_variables + n_randvars, )
            Standard errors of the estimated coefficients

        zvalues : numpy array, shape (n_variables + n_randvars, )
            Z-values for t-distribution of the estimated coefficients

        pvalues : numpy array, shape (n_variables + n_randvars, )
            P-values of the estimated coefficients

        loglikelihood : float
            Log-likelihood at the end of the estimation

        convergence : bool
            Whether convergence was reached during estimation

        total_iter : int
            Total number of iterations executed during estimation

        estim_time_sec : float
            Estimation time in seconds

        sample_size : int
            Number of samples used for estimation

        aic : float
            Akaike information criteria of the estimated model

        bic : float
            Bayesian information criteria of the estimated model
    """

    def __init__(self):
        """Init Function"""
        super(MixedLogit, self).__init__()
        self.rvidx = None  # Boolean index of random vars in X. True = rand var
        self.rvdist = None

    # X: (N, J, K)
    def fit(self, X, y, varnames=None, alts=None, isvars=None, transvars=None,
            transformation="boxcox", ids=None, weights=None, avail=None,
            randvars=None, panels=None, base_alt=None, fit_intercept=False,
            init_coeff=None, maxiter=2000, random_state=None, correlation=None,
            n_draws=1000, halton=True, minimise_func=None, verbose=1,
            batch_size=None, halton_opts=None, ftol=1e-7,
            gtol=1e-5, return_hess=True, return_grad=True, method="bfgs",
            save_fitted_params=True, mnl_init=True):
        """Fit Mixed Logit models.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_variables)
            Input data for explanatory variables in long format

        y : array-like, shape (n_samples,)
            Choices (outcome) in long format

        varnames : list, shape (n_variables,)
            Names of explanatory variables that must match the number and
            order of columns in ``X``

        alts : array-like, shape (n_samples,)
            Alternative indexes in long format or list of alternative names

        isvars : list
            Names of individual-specific variables in ``varnames``

        transvars: list, default=None
            Names of variables to apply transformation on

        transformation: string, default="boxcox"
            Name of transformation to apply on transvars

        ids : array-like, shape (n_samples,)
            Identifiers for choice situations in long format.

        weights : array-like, shape (n_variables,), default=None
            Weights for the choice situations in long format.

        avail: array-like, shape (n_samples,)
            Availability of alternatives for the choice situations. One when
            available or zero otherwise.

        randvars : dict
            Names (keys) and mixing distributions (values) of variables that
            have random parameters as coefficients. Possible mixing
            distributions are: ``'n'``: normal, ``'ln'``: lognormal,
            ``'u'``: uniform, ``'t'``: triangular, ``'tn'``: truncated normal

        panels : array-like, shape (n_samples,), default=None
            Identifiers in long format to create panels in combination with
            ``ids``

        base_alt : int, float or str, default=None
            Base alternative

        fit_intercept : bool, default=False
            Whether to include an intercept in the model.

        init_coeff : numpy array, shape (n_variables,), default=None
            Initial coefficients for estimation.

        maxiter : int, default=200
            Maximum number of iterations

        random_state : int, default=None
            Random seed for numpy random generator

        correlation: boolean or list, default=None
            If boolean finds correlation for all random (non trans) vars
            If list finds correlation between variables specified

        n_draws : int, default=1000
            Number of random draws to approximate the mixing distributions of
            the random coefficients

        halton : bool, default=True
            Whether the estimation uses halton draws.

        minimise_func : func, default=None
            Sets minimise function used.

        verbose : int, default=1
            Verbosity of messages to show during estimation. 0: No messages,
            1: Some messages, 2: All messages

        batch_size : int, default=None
            Size of batches of random draws used to avoid overflowing memory during computations.

        halton_opts : dict, default=None
            Options for generation of halton draws. The dictionary accepts the following options (keys):

                shuffle : bool, default=False
                    Whether the Halton draws should be shuffled

                drop : int, default=100
                    Number of initial Halton draws to discard to minimize correlations between Halton sequences

                primes : list
                    List of primes to be used as base for generation of Halton sequences.

        ftol : int, float, default=1e-7
            Sets the tol parameter in scipy.optimize.minimize - Tolerance for
            termination.

        gtol: int, float, default=1e-5
            Sets the gtol parameter in scipy.optimize.minimize(method="bfgs) -
            Gradient norm must be less than gtol before successful termination.

        return_hess : bool, default=True
            Calculate and return the gradient in _loglik_and_gradient

        return_grad : bool, default=True
            Calculate and return the gradient in _loglik_and_gradient

        method: string, default="bfgs"
            specify optimisation method

        mnl_init: bool, default=True
            Initialise coefficients with multinomial logit estimates.

        Returns
        -------
        None.
        """

        X, y, varnames, alts, isvars, transvars, ids, weights, \
            panels, avail = self._as_array(X, y, varnames, alts, isvars,
                                           transvars, ids, weights, panels,
                                           avail)
        # Handle array-like inputs by converting everything to numpy arrays
        self._validate_inputs(X, y, alts, varnames, isvars, ids, weights,
                              panels, base_alt, fit_intercept, maxiter)

        self._pre_fit(alts, varnames, isvars, transvars, base_alt,
                      fit_intercept, transformation, maxiter, panels,
                      correlation, randvars)

        X_original = X  # tmp fix... store to use in mnl_init later
        y_original = y

        batch_size = n_draws if batch_size is None else min(n_draws, batch_size)
        self.batch_size = batch_size

        self.randvarsdict = randvars  # random variables not transformed
        self.randvars = [x for x in self.randvars if x not in transvars]
        #  random variables that are transformed
        self.randtransvars = [x for x in transvars if (x in randvars) and
                              (x not in self.randvars)]
        self.fixedtransvars = [x for x in transvars if x not in
                               self.randtransvars]
        self.n_draws = n_draws
        self.gtol = gtol
        # divide the variables in varnames into fixed, fixed transformed,
        # random, random transformed by getting 4 index arrays
        # also for random and random transformed save the distributions
        # in a separate array
        self.fxidx, self.fxtransidx = [], []
        self.rvidx, self.rvdist = [], []
        self.rvtransidx, self.rvtransdist = [], []
        for var in self.varnames:
            if isvars is not None:
                if var in isvars:
                    continue
            if hasattr(randvars, 'keys') and var in randvars.keys():  # bug fix
                self.fxidx.append(False)
                self.fxtransidx.append(False)
                if var in self.randvars:
                    self.rvidx.append(True)
                    self.rvdist.append(randvars[var])
                    self.rvtransidx.append(False)
                else:
                    self.rvidx.append(False)
                    self.rvtransidx.append(True)
                    self.rvtransdist.append(randvars[var])
            else:
                self.rvidx.append(False)
                self.rvtransidx.append(False)
                self.rvdist.append(False)
                self.rvtransdist.append(False)
                if var in transvars:
                    self.fxtransidx.append(True)
                    self.fxidx.append(False)
                else:
                    self.fxtransidx.append(False)
                    self.fxidx.append(True)

        self.rvidx = np.array(self.rvidx)
        self.rvtransidx = np.array(self.rvtransidx)
        self.fxidx = np.array(self.fxidx)
        self.fxtransidx = np.array(self.fxtransidx)
        if random_state is not None:
            np.random.seed(random_state)
        X, y, panels = self._arrange_long_format(X, y, ids, alts, panels)
        X, Xnames = self._setup_design_matrix(X)
        self._model_specific_validations(randvars, Xnames)
        self.Xnames = Xnames  # saved here to use in LatentMixed...
        J, K, R = X.shape[1], X.shape[2], n_draws

        if self.transformation == "boxcox":
            self.trans_func = boxcox_transformation_mixed
            self.transform_deriv = boxcox_param_deriv_mixed

        R = n_draws
        if panels is not None:
            X, y, avail, panel_info = self._balance_panels(X, y, avail, panels)
            N, P = panel_info.shape
        else:
            N, P = X.shape[0], 1
            panel_info = np.ones((N, 1))

        X = X.reshape(N, P, J, K)
        y = y.reshape(N, P, J, 1)
        self.y = y
        self.panel_info = panel_info

        self.obs_prob = np.mean(np.mean(np.mean(y, axis=3), axis=1), axis=0)

        self.verbose = verbose

        self.return_grad = return_grad
        self.return_hess = return_hess
        if hasattr(method, 'lower'):
            self.method = method.lower()

        jac = True if self.return_grad else False

        self.total_fun_eval = 0

        #  reshape weights (using panel data if necessary)
        if weights is not None:
            weights = self._reshape_weights(weights, panels)

        if avail is not None:
            avail = self._reshape_avail(avail, panels)

        # Generate draws
        draws, drawstrans = self._generate_draws(self.N, R, halton)  # (N,Kr,R)

        # save draws for use in summary
        self.draws = draws
        self.drawstrans = drawstrans
        self.save_fitted_params = save_fitted_params
        # 2x Kftrans - mean and lambda, 3x Krtrans - mean, s.d., lambda
        # Kchol, Kbw - relate to random variables, non-transformed
        # Kchol - cholesky matrix, Kbw the s.d. for random vars
        n_coeff = self.Kf + self.Kr + self.Kchol + self.Kbw + 2*self.Kftrans +\
            3*self.Krtrans

        if mnl_init and init_coeff is None:
            # Initalise coefficients using a multinomial logit model
            mnl = MultinomialLogit()
            # varnames_mnl = [varname for varname in Xnames if 'sd' not in varname
            #                                               and 'chol' not in varname
            #                                               and 'lambda' not in varname]
            varnames_mnl = varnames
            mnl.fit(X_original,
                    y_original.flatten(),
                    varnames_mnl,
                    alts,
                    isvars,
                    transvars=transvars,
                    ids=ids,
                    weights=weights,
                    avail=avail,
                    base_alt=base_alt,
                    fit_intercept=fit_intercept,
                    verbose=0)

            init_coeff = mnl.coeff_
            # mnl estimates -> mxl needs to add stdev to random variables
            init_coeff = np.concatenate((init_coeff[:self.Kf+2*self.Kftrans+self.Kr],
                                         np.repeat(.1, self.Kchol + self.Kbw),
                                         init_coeff[self.Kf+2*self.Kftrans+self.Kr:self.Kf+2*self.Kftrans+self.Kr+self.Krtrans],
                                         ))
            if self.Krtrans:
                init_coeff = np.concatenate((init_coeff,
                                             np.repeat(.1, self.Krtrans),
                                             init_coeff[-self.Krtrans:]))
        if init_coeff is None:
            betas = np.repeat(.1, n_coeff)
        else:
            betas = init_coeff
            if len(init_coeff) != n_coeff and not hasattr(self, 'class_params_spec'):
                raise ValueError("The size of init_coeff must be: " + str(n_coeff))
        if dev.using_gpu:
            X, y = dev.to_gpu(X), dev.to_gpu(y)
            panel_info = dev.to_gpu(panel_info)
            draws = dev.to_gpu(draws)
            drawstrans = dev.to_gpu(drawstrans)
            if weights is not None:
                weights = dev.to_gpu(weights)
            if avail is not None:
                avail = dev.to_gpu(avail)
            if verbose > 0:
                print("Estimation with GPU processing enabled.")
        positive_bound = (0, 1e+30)
        any_bound = (-1e+30, 1e+30)
        # corr_bound = (-1, 1)
        lmda_bound = (-5, 1)

        # to be used with L-BFGS-B method, automatically generate bounds
        bound_dict = {
            "bf": (any_bound, self.Kf),
            "br_b": (any_bound, self.Kr),
            "chol": (any_bound, self.Kchol),
            "br_w": (positive_bound, self.Kr - self.correlationLength),
            "bf_trans": (any_bound, self.Kftrans),
            "flmbda": (lmda_bound, self.Kftrans),
            "br_trans_b": (any_bound, self.Krtrans),
            "br_trans_w": (any_bound, self.Krtrans),
            "rlmbda": (lmda_bound, self.Krtrans)
        }

        # list comrephension to add number of bounds for each variable type
        # bound[1][0] - the bound range
        # bound[1][1] - how many bounds to add
        bnds = [(bound[1][0],) * int(bound[1][1])
                for bound in bound_dict.items() if bound[1][1] > 0]
        # convert into appropriate format for L-BFGS-B
        bnds = [tuple(itertools.chain.from_iterable(bnds))][0]

        minimise_func = minimize if minimise_func is None else minimise_func

        optimizat_res = \
            minimise_func(
                self._loglik_gradient,
                betas,
                jac=jac,
                method=method,
                args=(X, y, panel_info, draws, drawstrans, weights,
                      avail, batch_size),
                tol=ftol,
                bounds=bnds if method == "L-BFGS-B" else None,
                options={
                    'gtol': gtol,
                    'maxiter': maxiter,
                    'disp': verbose > 0,
                }
            )
        if hasattr(self, 'method') and self.method == "L-BFGS-B":
            # calculate hessian by finite differences at minimisation point
            def _hessian(x, func, eps=1e-8):
                N = x.size
                h = np.zeros((N, N))
                df_0 = func(x)[1]
                for i in np.arange(N):
                    xx0 = 1 * x[i]
                    x[i] = xx0 + eps
                    df_1 = func(x)[1]
                    h[i, :] = (df_1 - df_0)/eps
                    x[i] = xx0
                return h

            def _get_loglik_gradient(betas):
                res = self._loglik_gradient(betas, X, y, panel_info, draws,
                                            drawstrans, weights, avail)
                return res

            H = _hessian(optimizat_res['x'], _get_loglik_gradient)
            H = np.linalg.inv(H)
            optimizat_res['hess_inv'] = H

        if self.save_fitted_params:
            self._save_fitted_params(self.y, self.p, panel_info, self.Br)

        # save predicted and observed probabilities to display in summary
        try:
            if 'is_latent_class' in optimizat_res.keys():
                pass
            else:
                p = self._compute_probabilities(optimizat_res['x'],
                                                X,
                                                panel_info,
                                                draws,
                                                drawstrans,
                                                avail,
                                                self.var_list,
                                                self.chol_mat)

                self.choice_pred_prob = np.mean(p, axis=3)
                self.ind_pred_prob = np.mean(self.choice_pred_prob, axis=1)
                self.pred_prob = np.mean(self.ind_pred_prob, axis=0)
                self.prob_full = p
        except Exception:
            pass
        self._post_fit(optimizat_res, Xnames, N, verbose=verbose)

    def _reshape_weights(self, weights, panels):
        """Logic to reshape weights according to panels."""
        N, P, J = self.N, self.P, self.J
        weights = weights*(N/np.sum(weights))  # Normalize weights
        if panels is not None:
            # copied logic from _balance_panels function
            _, p_obs = np.unique(panels, return_counts=True)
            p_obs = (p_obs/J).astype(int)
            weights_temp = np.zeros((N, P, J))
            cum_p = 0
            for n, p in enumerate(p_obs):
                weights_temp[n, 0:p, :] = weights[cum_p:cum_p+(p*J)].reshape((1, p, J))
                cum_p += p
            weights = weights_temp.reshape(N, P, J)
        else:
            weights = weights.reshape(N, J)
        return weights

    def _reshape_avail(self, avail, panels):
        """Logic to reshape avail according to panels."""
        N, P, J = self.N, self.P, self.J
        if panels is not None:
            # copied logic from _balance_panels function
            _, p_obs = np.unique(panels, return_counts=True)
            p_obs = (p_obs/J).astype(int)
            avail_temp = np.zeros((N, P, J))
            cum_p = 0
            for n, p in enumerate(p_obs):
                avail_temp[n, 0:p, :] = avail[cum_p:cum_p+(p*J)].reshape((1, p, J))
                cum_p += p
            avail = avail_temp.reshape(N, P, J)
        else:
            avail = avail.reshape(N, J)
        return avail

    def _save_fitted_params(self, y, p, panel_info, Br):
        # TODO: could improve ... look at xlogit
        # calculate fitted parameter values
        if dev.using_gpu:
            y, p = dev.to_gpu(y), dev.to_gpu(p)
        pch = np.sum(y*p, axis=2)  # (N, P, R)
        pch = self._prob_product_across_panels(pch, panel_info)
        # Thresholds to avoid divide by zero warnings
        pch[pch == 0] = min_comp_val

        pch2 = np.divide(pch, np.sum(pch, axis=1)[:, None])  # pch/rowsum(pch)

        pch2 = pch2.flatten()
        temp_br = np.zeros((self.N*self.batch_size, self.Kr))
        if dev.using_gpu:
            temp_br = dev.to_gpu(temp_br)
        for i in range(self.N):
            for k in range(self.Kr):
                temp_br[(i*self.batch_size):((i+1)*self.batch_size), k] = Br[i, k, :]

        Br = temp_br

        pch2 = np.multiply(pch2[:, None], Br)
        # if not dev.using_gpu:
        pch2_res = np.zeros((self.N, self.Kr))
        pch2_sd_test = np.zeros((self.N, self.Kr))
        if dev.using_gpu:
            pch = dev.to_gpu(pch)
            pch2_res = dev.to_gpu(pch2_res)
            pch2_sd_test = dev.to_gpu(pch2_sd_test)
        for i in range(self.N):
            # print('pch2[i*self.n_draws:(i+1)*self.n_draws, :]', pch2[i*self.n_draws:(i+1)*self.n_draws, :])
            pch2_res[i, :] = np.sum(pch2[i*self.batch_size:(i+1)*self.batch_size, :], axis=0)
            pch2_sd_test[i, :] = np.std(pch2[i*self.batch_size:(i+1)*self.batch_size, :],
                                        axis=0)

        self.pch2_res = pch2_res
        self.pch2_sd_test = pch2_sd_test

    def _construct_chol_mat(self, chol, Br_w, Brtrans_w):
        # creating cholesky matrix for the variance-covariance matrix
        # all random variables not included in correlation will only
        # have their standard deviation computed
        # NOTE: fairly poorly written function, made to patch bugs
        chol_mat = np.zeros((self.correlationLength, self.correlationLength))
        indices = np.tril_indices(self.correlationLength)
        if dev.using_gpu:
            chol = dev.to_cpu(chol)
            chol_mat[indices] = chol  # TODO? Better
        else:
            chol_mat[indices] = chol
        Kr_all = self.Kr + self.Krtrans
        chol_mat_temp = np.zeros((Kr_all, Kr_all))

        # TODO: Structure ... Kr first, Krtrans last, fill in for correlations
        # TODO: could do better
        rv_count = 0
        rv_count_all = 0
        rv_trans_count = 0
        rv_trans_count_all = 0
        chol_count = 0
        corr_indices = []

        # TODO: another bugfix
        # Know beforehand to order rvtrans correctly
        num_corr_rvtrans = 0
        for ii, var in enumerate(self.varnames):
            if self.rvtransidx[ii]:
                if hasattr(self, 'correlation') and self.correlation:
                    if hasattr(self.correlation, 'append'):
                        if var in self.correlation:
                            num_corr_rvtrans += 1
                    # else:
                    #     num_corr_rvtrans += 1

        num_rvtrans_total = self.Krtrans + num_corr_rvtrans

        offset_varnames = []
        if hasattr(self, 'Xnames'):
            offset_varnames = self.Xnames
        else:
            offset_varnames = self.varnames
        inter_offset = 0
        if not hasattr(self, 'class_params_spec'):
            inter_offset = len([x for x in offset_varnames if 'inter' in x]) - 1
        if inter_offset < 0:
            inter_offset = 0
        for ii, var in enumerate(self.varnames):  # TODO: BUGFIX
            ii_offset = ii + inter_offset
            is_correlated = False
            if hasattr(self, 'correlation') and self.correlation:
                if hasattr(self.correlation, 'append'):
                    if var in self.correlation:
                        is_correlated = True
                else:
                    is_correlated = True
            if self.rvidx[ii_offset]:
                rv_val = chol[chol_count] if is_correlated else Br_w[rv_count]
                chol_mat_temp[rv_count_all, rv_count_all] = rv_val
                rv_count_all += 1
                if is_correlated:
                    chol_count += 1
                else:
                    rv_count += 1
            if self.rvtransidx[ii_offset]:
                if hasattr(self, 'correlation') and isinstance(self.correlation, bool):  # BUG FIX
                    is_correlated = False
                rv_val = chol[chol_count] if is_correlated else Brtrans_w[rv_trans_count]
                chol_mat_temp[-num_rvtrans_total + rv_trans_count_all, -num_rvtrans_total + rv_trans_count_all] = rv_val
                rv_trans_count_all += 1
                if is_correlated:
                    chol_count += 1
                else:
                    rv_trans_count += 1
            if hasattr(self, 'correlation') and self.correlation:
                if hasattr(self.correlation, 'append'):
                    if var in self.correlation:
                        if self.rvidx[ii_offset]:
                            corr_indices.append(rv_count_all - 1)
                        else:
                            corr_indices.append(Kr_all - num_rvtrans_total + rv_trans_count_all - 1)   # TODO i think

        if hasattr(self, 'correlation') and isinstance(self.correlation, bool) and self.correlation:
            corr_pairs = list(itertools.combinations(np.arange(self.Kr), 2)) + [(i, i) for i in range(self.Kr)]
        else:
            corr_pairs = list(itertools.combinations(corr_indices, 2)) + [(idx, idx) for ii, idx in enumerate(corr_indices)]

        reversed_corr_pairs = [tuple(reversed(pair)) for ii, pair in enumerate(corr_pairs)]
        reversed_corr_pairs.sort(key=lambda x: x[0])

        chol_count = 0
        for _, corr_pair in enumerate(reversed_corr_pairs):
            # lower cholesky matrix
            chol_mat_temp[corr_pair] = chol[chol_count]
            chol_count += 1
        # # add non-correlated random variables to cholesky matrix
        # num_rand_noncorr = sel f.Kr - self.correlationLength
        # for i in range(num_rand_noncorr):
        #     chol_mat_temp[i, i] = Br_w[i]
        # # add non-correlated transformed random variables to cholesky matrix
        # for i in range(self.Krtrans):
        #     chol_mat_temp[i + num_rand_noncorr, i + num_rand_noncorr] = Brtrans_w[i]

        # if self.correlationLength > 0:
        #     chol_mat_temp[-self.correlationLength:, -self.correlationLength:] = \
        #         chol_mat

        chol_mat = chol_mat_temp

        return chol_mat

    def _compute_probabilities(self, betas, X, panel_info, draws, drawstrans,
                               avail, var_list, chol_mat):
        """Compute choice probabilities for each alternative."""
        # Creating random coeffs using Br_b, cholesky matrix and random draws
        # Estimating the linear utility specification (U = sum of Xb)

        Bf, Br_b, chol, Br_w, Bftrans, flmbda, Brtrans_b, Brtrans_w, rlmda = \
            var_list.values()
        if dev.using_gpu:
            Bf, Br_b, chol, Br_w, Bftrans, flmbda, Brtrans_b, \
                Brtrans_w, rlmda = \
                dev.to_gpu(Bf), dev.to_gpu(Br_b), dev.to_gpu(chol), \
                dev.to_gpu(Br_w), dev.to_gpu(Bftrans), dev.to_gpu(flmbda), \
                dev.to_gpu(Brtrans_b), dev.to_gpu(Brtrans_w), dev.to_gpu(rlmda)

        Xf = X[:, :, :, self.fxidx]
        Xftrans = X[:, :, :, self.fxtransidx]
        Xr = X[:, :, :, self.rvidx]
        Xrtrans = X[:, :, :, self.rvtransidx]

        XBf = np.zeros((self.N, self.P, self.J))
        XBr = np.zeros((self.N, self.P, self.J, self.batch_size))
        V = np.zeros((self.N, self.P, self.J, self.batch_size))
        Br = None
        if dev.using_gpu:
            XBf = dev.to_gpu(XBf)
            XBr = dev.to_gpu(XBr)
            V = dev.to_gpu(V)
        if self.Kf != 0:
            XBf = dev.cust_einsum('npjk,k -> npj', Xf, Bf)

        if self.Kr != 0:
            Br = Br_b[None, :, None] + dev.np.matmul(chol_mat[:self.Kr, :self.Kr], draws)
            Br = self._apply_distribution(Br, self.rvdist)

            self.Br = Br  # save Br to use later
            XBr = dev.cust_einsum('npjk,nkr -> npjr', Xr, Br)  # (N, P, J, R)
            V = XBf[:, :, :, None] + XBr

        #  transformations for variables with fixed coeffs
        if self.Kftrans != 0:
            Xftrans_lmda = self.trans_func(Xftrans, flmbda)
            Xftrans_lmda[np.where(Xftrans_lmda < -max_comp_val)] = -max_comp_val
            Xftrans_lmda[np.where(Xftrans_lmda > max_comp_val)] = max_comp_val
            # Estimating the linear utility specificiation (U = sum XB)
            Xbf_trans = dev.cust_einsum('npjk,k -> npj', Xftrans_lmda, Bftrans)
            # combining utilities
            V += Xbf_trans[:, :, :, None]

        # transformations for variables with random coeffs
        if self.Krtrans != 0:
            # creating the random coeffs
            Brtrans = Brtrans_b[None, :, None] + \
                    drawstrans[:, 0:self.Krtrans, :] * Brtrans_w[None, :, None]
            Brtrans = self._apply_distribution(Brtrans, self.rvtransdist)
            self.Brtrans = Brtrans  # saving for later use
            # applying transformation
            Xrtrans_lmda = self.trans_func(Xrtrans, rlmda)
            Xrtrans_lmda[np.where(Xrtrans_lmda > max_comp_val)] = max_comp_val
            Xrtrans_lmda[np.where(Xrtrans_lmda < -max_comp_val)] = -max_comp_val

            Xbr_trans = dev.cust_einsum('npjk,nkr -> npjr', Xrtrans_lmda, Brtrans)  # (N, P, J, R)
            # combining utilities
            V += Xbr_trans  # (N, P, J, R)

        # Thresholds to avoid overflow warnings
        V[np.where(V > max_exp_val)] = max_exp_val
        V[np.where(V < -max_exp_val)] = -max_exp_val
        eV = dev.np.exp(V)

        # Exponent of the utility function for the logit formula
        if avail is not None:
            if self.panels is not None:
                eV = eV*avail[:, :, :, None]  # Acommodate availablity of alts with panels
            else:
                eV = eV*avail[:, None, :, None]  # Acommodate availablity of alts.

        sum_eV = dev.np.sum(eV, axis=2, keepdims=True)
        sum_eV[np.where(sum_eV < min_comp_val)] = min_comp_val
        p = np.divide(eV, sum_eV, out=np.zeros_like(eV))
        if self.save_fitted_params:
            self.p = p  # save

        return p

    def _loglik_gradient(self, betas, X, y, panel_info, draws, drawstrans,
                         weights, avail, batch_size):
        """Compute the log-likelihood and gradient.

        Fixed and random parameters are handled separately to
        speed up the estimation and the results are concatenated.
        """
        # Segregating initial values to fixed betas (Bf),
        # random beta means (Br_b)
        # for both non-transformed and transformed variables
        # and random beta cholesky factors (chol)
        self.betas = betas  # save to display later

        if dev.using_gpu:
            betas = dev.to_gpu(betas)

        beta_segment_names = ["Bf", "Br_b", "chol", "Br_w", "Bftrans",
                              "flmbda", "Brtrans_b", "Brtrans_w", "rlmda"]
        var_list = dict()
        # number of parameters for each corresponding segment
        iterations = [self.Kf, self.Kr, self.Kchol, self.Kbw, self.Kftrans,
                      self.Kftrans, self.Krtrans, self.Krtrans, self.Krtrans]

        i = 0
        for count, iteration in enumerate(iterations):
            prev_index = i
            i = int(i + iteration)
            var_list[beta_segment_names[count]] = betas[prev_index:i]

        Bf, Br_b, chol, Br_w, Bftrans, flmbda, Brtrans_b, Brtrans_w, rlmda = \
            var_list.values()
        if dev.using_gpu:
            Bf, Br_b, chol, Br_w, Bftrans, flmbda, Brtrans_b, \
                Brtrans_w, rlmda = \
                dev.to_gpu(Bf), dev.to_gpu(Br_b), dev.to_gpu(chol), \
                dev.to_gpu(Br_w), dev.to_gpu(Bftrans), dev.to_gpu(flmbda), \
                dev.to_gpu(Brtrans_b), dev.to_gpu(Brtrans_w), dev.to_gpu(rlmda)

        chol_mat = self._construct_chol_mat(chol, Br_w, Brtrans_w)

        if dev.using_gpu:
            chol_mat = dev.to_gpu(chol_mat)

        omega = dev.np.matmul(chol_mat, np.transpose(chol_mat))
        corr_mat = np.zeros_like(chol_mat)

        standard_devs = np.sqrt(np.diag(omega))
        K = len(standard_devs)
        for i in range(K):
            for j in range(K):
                corr_mat[i, j] = omega[i, j] / (standard_devs[i] * standard_devs[j])
        self.omega = omega
        self.stdevs = standard_devs
        self.corr_mat = corr_mat

        R = self.n_draws
        n_batches = R//batch_size + (1 if R % batch_size != 0 else 0)
        N = self.N
        _, gr_b, gr_w, pch = np.zeros((N, self.Kf)), np.zeros((N, self.Kr)), np.zeros((N, self.Kr)), []  # Batch data

        g_all = np.zeros((N, len(betas)))

        for batch in range(n_batches):
            draws_batch = draws[:, :, batch*batch_size: batch*batch_size + batch_size]
            drawstrans_batch = drawstrans[:, :, batch*batch_size: batch*batch_size + batch_size]
            if dev.using_gpu:
                draws_batch = dev.to_gpu(draws_batch)

            self.chol_mat = chol_mat
            self.var_list = var_list

            p = self._compute_probabilities(betas, X, panel_info, draws_batch, drawstrans_batch,
                                            avail, var_list, chol_mat)

            # Joint probability estimation for panels data
            pch_batch = np.sum(y*p, axis=2)  # (N, P, R)
            pch_batch = self._prob_product_across_panels(pch_batch, panel_info)
            # Thresholds to avoid divide by zero warnings
            pch_batch[pch_batch == 0] = min_comp_val

            # Observed probability minus predicted probability
            ymp = y - p  # (N, P, J, R)

            Xf = X[:, :, :, self.fxidx]
            Xftrans = X[:, :, :, self.fxtransidx]
            Xr = X[:, :, :, self.rvidx]
            Xrtrans = X[:, :, :, self.rvtransidx]

            # For fixed params
            # gradient = (Obs prob. minus predicted probability) * obs. var
            g = np.array([])
            if self.Kf != 0:
                g = dev.cust_einsum('npjr,npjk -> nkr', ymp, Xf)
                g = (g*pch_batch[:, None, :]).mean(axis=2)

            # For random params w/ untransformed vars, two gradients will be
            # estimated: one for the mean and one for the s.d.
            # for mean: gr_b = (Obs. prob. minus pred. prob.)  * obs. var
            # for s.d.: gr_b = (Obs. prob. minus pred. prob.)  * obs. var * rand draw
            # if random coef. is lognormally dist:
            # gr_b = (obs. prob minus pred. prob.) * obs. var. * rand draw * der(R.V.)

            if self.Kr != 0:
                der = self._compute_derivatives(betas, draws_batch, chol_mat=chol_mat, betas_random=self.Br)
                gr_b = dev.cust_einsum('npjr,npjk -> nkr', ymp, Xr)*der  # (N, Kr, R)
                # For correlation parameters
                # for s.d.: gr_w = (Obs prob. minus predicted probability) * obs. var * random draw
                draws_tril_idx = np.array([j
                                           for i in range(self.correlationLength)
                                           for j in range(i+1)])  # varnames pos.
                X_tril_idx = np.array([i
                                       for i in range(self.correlationLength)
                                       for j in range(i+1)])
                # Find the s.d. for random variables that are not correlated
                range_var = [x for x in
                             range(self.correlationLength, self.Kr)]
                range_var = sorted(range_var)
                draws_tril_idx = np.array(np.concatenate((draws_tril_idx,
                                                          range_var)))
                X_tril_idx = np.array(np.concatenate((X_tril_idx, range_var)))
                draws_tril_idx = draws_tril_idx.astype(int)
                X_tril_idx = X_tril_idx.astype(int)
                gr_w = gr_b[:, X_tril_idx, :]*draws_batch[:, draws_tril_idx, :]  # (N,P,Kr,R)
                gr_b = (gr_b*pch_batch[:, None, :]).mean(axis=2)  # (N,Kr)
                gr_w = (gr_w*pch_batch[:, None, :]).mean(axis=2)  # (N,Kr)

                # Gradient for fixed and random params
                g = np.concatenate((g, gr_b, gr_w), axis=1) if g.size \
                    else np.concatenate((gr_b, gr_w), axis=1)

            # For Box-Cox vars
            if len(self.transvars) > 0:
                if self.Kftrans:  # with fixed params
                    Xftrans_lmda = self.trans_func(Xftrans, flmbda)
                    Xftrans_lmda[np.where(Xftrans_lmda < -max_comp_val)] = -max_comp_val
                    Xftrans_lmda[np.where(Xftrans_lmda > max_comp_val)] = max_comp_val

                    gftrans = dev.cust_einsum('npjr,npjk -> nkr', ymp,
                                              Xftrans_lmda)  # (N, Kf, R)
                    # for the lambda param
                    der_Xftrans_lmda = self.transform_deriv(Xftrans, flmbda)
                    der_Xftrans_lmda[np.where(der_Xftrans_lmda > max_comp_val)] = max_comp_val
                    der_Xftrans_lmda[np.where(der_Xftrans_lmda < -max_comp_val)] = -max_comp_val
                    der_Xftrans_lmda[np.isnan(der_Xftrans_lmda)] = min_comp_val
                    der_Xbftrans = dev.np.einsum('npjk,k -> njk', der_Xftrans_lmda,
                                             Bftrans)

                    gftrans_lmda = dev.np.einsum('npjr,njk -> nkr', ymp,
                                             der_Xbftrans)
                    gftrans = (gftrans*pch_batch[:, None, :]).mean(axis=2)
                    gftrans_lmda = (gftrans_lmda*pch_batch[:, None, :]).mean(axis=2)
                    g = np.concatenate((g, gftrans, gftrans_lmda), axis=1) if g.size \
                        else np.concatenate((gftrans, gftrans_lmda), axis=1)

                if self.Krtrans:
                    # for rand parameters
                    # for mean: (obs prob. min pred. prob)*obs var * deriv rand coef
                    # if rand coef is lognormally distributed:
                    # gr_b = (obs prob minus pred. prob) * obs. var * rand draw * der(RV)
                    temp_chol = chol_mat if chol_mat.size != 0 else np.diag(Brtrans_w)
                    dertrans = self._compute_derivatives(betas,
                                                         draws=drawstrans_batch,
                                                         dist=self.rvtransdist,
                                                         chol_mat=temp_chol,
                                                         K=self.Krtrans,
                                                         trans=True,
                                                         betas_random=self.Brtrans
                                                         )
                    Xrtrans_lmda = self.trans_func(Xrtrans, rlmda)

                    Brtrans = Brtrans_b[None, :, None] + \
                        drawstrans[:, 0:self.Krtrans, :] * \
                        Brtrans_w[None, :, None]
                    Brtrans = self._apply_distribution(Brtrans,
                                                       self.rvtransdist)

                    grtrans_b = dev.cust_einsum('npjr,npjk -> nkr', ymp,
                                                Xrtrans_lmda)*dertrans
                    # for s.d. (obs - pred) * obs var * der rand coef * rd draw
                    grtrans_w = dev.cust_einsum('npjr,npjk -> nkr', ymp,
                                                Xrtrans_lmda) * dertrans * \
                        drawstrans_batch
                    # for the lambda param
                    # gradient = (obs - pred) * deriv x_lambda * beta
                    der_Xrtrans_lmda = self.transform_deriv(Xrtrans, rlmda)
                    der_Xrtrans_lmda[np.where(der_Xrtrans_lmda > max_comp_val)] = max_comp_val
                    der_Xrtrans_lmda[np.where(der_Xrtrans_lmda < -max_comp_val)] = max_comp_val
                    # TODO? KEEP an eye out ... 4, 3, 5 ...  no cust_einsum
                    der_Xbrtrans = dev.np.einsum('npjk, nkr -> npjkr', der_Xrtrans_lmda,
                                             Brtrans)  # (N, P, J, K, R)
                    grtrans_lmda = dev.np.einsum('npjr, npjkr -> nkr', ymp, der_Xbrtrans)  # (N, Krtrans, R)
                    grtrans_b = (grtrans_b*pch_batch[:, None, :]).mean(axis=2)  # (N,Kr)
                    grtrans_w = (grtrans_w*pch_batch[:, None, :]).mean(axis=2)  # (N,Kr)
                    grtrans_lmda = (grtrans_lmda*pch_batch[:, None, :]).mean(axis=2)  # (N,Kr)
                    g = np.concatenate((g, grtrans_b, grtrans_w, grtrans_lmda), axis=1) if g.size \
                        else np.concatenate((grtrans_b, grtrans_w, grtrans_lmda), axis=1)

            if dev.using_gpu:
                g = dev.to_cpu(g)
                pch_batch = dev.to_cpu(pch_batch)

            pch.append(pch_batch)

            g_all += g

        pch = np.concatenate(pch, axis=-1)
        lik = pch.mean(axis=1)  # (N,)
        loglik = np.log(lik)
        if weights is not None:
            loglik = loglik*weights
        loglik = loglik.sum()

        self.total_fun_eval += 1
        g = g_all

        g = g/lik[:, None]
        # updated gradient
        if weights is not None:
            if weights.ndim == 1:
                g = np.transpose(np.transpose(g) * weights)
            if weights.ndim == 2:
                g = np.transpose(np.transpose(g)*weights[:, 0])
            if weights.ndim == 3:  # (3 dim for panel data)
                g = np.transpose(np.transpose(g)*weights[:, 0, 0])
        g = np.sum(g, axis=0)/n_batches  # (K, )
        # log-lik
        self.gtol_res = np.linalg.norm(g, ord=np.inf)

        result = (-loglik)

        if self.return_grad:
            result = (-loglik, -g)

        return result

    def validation_loglik(self, validation_X, validation_Y, betas=None,
                          avail=None, weights=None, panels=None):
        """Computes the log-likelihood on the validation set.

        Using estimated parameters, computes the log-likelihood on the
        validation set.
        """
        if panels is not None:
            N = len(np.unique(panels))
        else:
            N = self.N
        validation_X, Xnames = self._setup_design_matrix(validation_X)
        self.N = N

        betas = betas if betas is not None else self.coeff_
        if panels is not None:
            validation_X, validation_Y, avail, panel_info =   \
                self._balance_panels(validation_X, validation_Y, avail, panels)
        else:
            panel_info = np.ones((self.N, 1))

        if avail is not None:
            avail = self._reshape_avail(avail, panels)

        if weights is not None:
            weights = self._reshape_weights(weights, panels)

        validation_X = validation_X.reshape((N, self.P, self.J, -1))
        validation_Y = validation_Y.reshape((N, self.P, self.J, -1))
        draws, drawstrans = self._generate_draws(N, self.n_draws)  # (N,Kr,R)

        if dev.using_gpu:
            validation_X, validation_Y = dev.to_gpu(validation_X), dev.to_gpu(validation_Y)
            panel_info = dev.to_gpu(panel_info)
            draws = dev.to_gpu(draws)
            drawstrans = dev.to_gpu(drawstrans)
            if weights is not None:
                weights = dev.to_gpu(weights)
            if avail is not None:
                avail = dev.to_gpu(avail)

        self.y = validation_Y

        res = self._loglik_gradient(betas, validation_X, validation_Y,
                                    avail=avail, weights=weights,
                                    panel_info=panel_info,
                                    draws=draws, drawstrans=drawstrans,
                                    batch_size=self.batch_size
                                    )
        loglik = -res[0]
        return loglik

    def _prob_product_across_panels(self, pch, panel_info):
        if not np.all(panel_info):  # If panels unbalanced. Not all ones
            idx = panel_info == 0
            for i in range(pch.shape[2]):
                pch[:, :, i][idx] = 1  # Multiply by one when unbalanced
        pch = pch.prod(axis=1)  # (N,R)
        pch[pch == 0] = min_comp_val
        return pch  # (N,R)

    def _apply_distribution(self, betas_random, index=None, draws=None):
        """Apply the mixing distribution to the random betas."""
        index = index if (index is not None) else self.rvdist
        for k, dist in enumerate(index):
            if dist == 'ln':
                betas_random[:, k, :] = dev.np.exp(betas_random[:, k, :])
            elif dist == 'tn':
                betas_random[:, k, :] = betas_random[:, k, :] *\
                    (betas_random[:, k, :] > 0)
        return betas_random

    def _balance_panels(self, X, y, avail, panels):
        """Balance panels if necessary and produce a new version of X and y.

        If panels are already balanced, the same X and y are returned. This
        also returns panel_info, which keeps track of the panels that needed
        balancing.
        """
        _, J, K = X.shape
        _, p_obs = np.unique(panels, return_counts=True)
        p_obs = (p_obs/J).astype(int)
        N = len(p_obs)  # This is the new N after accounting for panels
        P = np.max(p_obs)  # panels length for all records
        if not np.all(p_obs[0] == p_obs):  # Balancing needed
            y = y.reshape(X.shape[0], J, 1) if y is not None else None
            avail = avail.reshape(X.shape[0], J, 1) if avail is not None else None
            Xbal, ybal, availbal = np.zeros((N*P, J, K)), np.zeros((N*P, J, 1)), np.zeros((N*P, J, 1))
            panel_info = np.zeros((N, P))
            cum_p = 0  # Cumulative sum of n_obs at every iteration
            for n, p in enumerate(p_obs):
                # Copy data from original to balanced version
                Xbal[n*P:n*P + p, :, :] = X[cum_p:cum_p + p, :, :]
                ybal[n*P:n*P + p, :, :] = y[cum_p:cum_p + p, :, :] if y is not None else None  # TODO? predict mode in xlogit?
                availbal[n*P:n*P + p, :, :] = avail[cum_p:cum_p + p, :, :] if avail is not None else None
                panel_info[n, :p] = np.ones(p)
                cum_p += p
        else:  # No balancing needed
            Xbal, ybal, availbal = X, y, avail
            panel_info = np.ones((N, P))
        ybal = ybal if y is not None else None
        availbal = availbal if avail is not None else None
        return Xbal, ybal, availbal, panel_info

    def _compute_derivatives(self, betas, draws, dist=None, K=None,
                             chol_mat=None, trans=False, betas_random=None):
        """Compute the derivatives based on the mixing distributions."""
        N, R = draws.shape[0], draws.shape[2]
        Kr = K if K else self.Kr

        der = dev.np.ones((N, Kr, R))
        dist = dist if dist else self.rvdist
        if any(set(dist).intersection(['ln', 'tn'])):  # If any ln or tn
            for k, dist_k in enumerate(dist):
                if dist_k == 'ln':
                    der[:, k, :] = betas_random[:, k, :]
                elif dist_k == 'tn':
                    der[:, k, :] = 1*(betas_random[:, k, :] > 0)
        return der

    def _transform_betas(self, betas, draws, index=None, trans=False, chol_mat=None):
        """Compute the products between the betas and the random coefficients.

        This method also applies the associated mixing distributions.
        """
        if trans:
            br_mean = betas[-3*self.Krtrans:-2*self.Krtrans]  # get pos from end array
            br_sd = betas[-2*self.Krtrans:-self.Krtrans]
            betas_random = br_mean[None, :, None] + draws*br_sd[None, :, None]
            betas_fixed = []
            betas_random = self._apply_distribution(betas_random, index,
                                                    draws=draws)
        else:
            # Extract coeffiecients from betas array
            betas_fixed = betas[0:self.Kf]  # First Kf positions
            br_mean = betas[self.Kf:self.Kf+self.Kr]

            betas_random = br_mean[None, :, None] + dev.np.matmul(chol_mat, draws)
            betas_random = self._apply_distribution(betas_random, index,
                                                    draws=draws)
        return betas_fixed, betas_random

    def _generate_draws(self, sample_size, n_draws, halton=True, chol_mat=None):
        """Generate draws based on the given mixing distributions."""
        draws = drawstrans = []
        if halton:
            if self.randvars:
                draws = self._generate_halton_draws(sample_size, n_draws,
                                                    np.sum(self.rvidx))
            if self.randtransvars:
                drawstrans = \
                    self._generate_halton_draws(sample_size,
                                                n_draws,
                                                np.sum(self.rvtransidx))
        else:
            if self.randvars:
                draws = self._get_random_draws(sample_size, n_draws,
                                               np.sum(self.rvidx))
            if self.randtransvars:
                drawstrans = self._get_random_draws(sample_size, n_draws,
                                                    np.sum(self.rvtransidx))
        # remove False to allow better enumeration
        self.rvdist = [x for x in self.rvdist if x is not False]
        for k, dist in enumerate(self.rvdist):
            if dist in ['n', 'ln', 'tn']:  # Normal based
                draws[:, k, :] = ss.norm.ppf(draws[:, k, :])
            elif dist == 't':  # Triangular
                draws_k = draws[:, k, :]
                draws[:, k, :] = (np.sqrt(2*draws_k) - 1)*(draws_k <= .5) +\
                    (1 - np.sqrt(2*(1 - draws_k)))*(draws_k > .5)
            elif dist == 'u':  # Uniform
                draws[:, k, :] = 2*draws[:, k, :] - 1

        # remove False to allow better enumeration
        self.rvtransdist = [x for x in self.rvtransdist if x is not False]
        for k, dist in enumerate(self.rvtransdist):
            if dist in ['n', 'ln', 'tn']:  # Normal based
                drawstrans[:, k, :] = ss.norm.ppf(drawstrans[:, k, :])
            elif dist == 't':  # Triangular
                draws_k = drawstrans[:, k, :]
                drawstrans[:, k, :] = (np.sqrt(2*draws_k) - 1)*(draws_k <= .5) +\
                    (1 - np.sqrt(2*(1 - draws_k)))*(draws_k > .5)
            elif dist == 'u':  # Uniform
                drawstrans[:, k, :] = 2*drawstrans[:, k, :] - 1
        draws = np.atleast_3d(draws)
        drawstrans = np.atleast_3d(drawstrans)
        return draws, drawstrans  # (N,Kr,R)

    def _get_random_draws(self, sample_size, n_draws, n_vars):
        """Generate random uniform draws between 0 and 1."""
        return np.random.uniform(size=(sample_size, n_vars, n_draws))

    def _generate_halton_draws(self, sample_size, n_draws, n_vars, shuffled=False, drop=100, primes=None):
        """Generate Halton draws for multiple random variables using different primes as base."""
        if primes is None:
            primes = [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47,
                      53, 59, 61, 71, 73, 79, 83, 89, 97, 101, 103, 107, 109,
                      113, 127, 131, 137, 139, 149, 151, 157, 163, 167, 173,
                      179, 181, 191, 193, 197, 199, 211, 223, 227, 229, 233,
                      239, 241, 251, 257, 263, 269, 271, 277, 281, 283, 293,
                      307, 311]

        def halton_seq(length, prime=3, shuffled=False, drop=100):
            """Generates a halton sequence while handling memory efficiently.

            Memory is efficiently handled by creating a single array ``seq`` that is iteratively filled without using
            intermidiate arrays.
            """
            req_length = length + drop
            seq = np.empty(req_length)
            seq[0] = 0
            seq_idx = 1
            t = 1
            while seq_idx < req_length:
                d = 1/prime**t
                seq_size = seq_idx
                i = 1
                while i < prime and seq_idx < req_length:
                    max_seq = min(req_length - seq_idx, seq_size)
                    seq[seq_idx: seq_idx+max_seq] = seq[:max_seq] + d*i
                    seq_idx += max_seq
                    i += 1
                t += 1
            seq = seq[drop:length+drop]
            if shuffled:
                np.random.shuffle(seq)
            return seq

        draws = [halton_seq(sample_size*n_draws, prime=primes[i % len(primes)],
                            shuffled=shuffled, drop=drop).reshape(sample_size, n_draws) for i in range(n_vars)]
        draws = np.stack(draws, axis=1)
        return draws  # (N,Kr,R)

    def _model_specific_validations(self, randvars, Xnames):
        """Conduct validations specific for mixed logit models."""
        if randvars is None:
            raise ValueError("The randvars parameter is required for Mixed "
                             "Logit estimation")
        if not set(randvars.keys()).issubset(Xnames):
            raise ValueError("Some variable names in randvars were not found "
                             "in the list of variable names")
        if not set(randvars.values()).issubset(["n", "ln", "t", "tn", "u"]):
            raise ValueError("Wrong mixing distribution found in randvars. "
                             "Accepted distrubtions are n, ln, t, u, tn")

    def summary(self):
        """Show estimation results in console."""
        super(MixedLogit, self).summary()

    def _compute_null_loglik(self):
        null_p = (1/self.J)*np.ones_like(self.y)
        p = np.sum(self.y*null_p, axis=2)
        null_loglik = np.log(self._prob_product_across_panels(p,
                                                              self.panel_info))
        null_loglik = np.sum(null_loglik)
        null_loglik = -2 * null_loglik
        return null_loglik

    @staticmethod
    def check_if_gpu_available():
        """Check if GPU processing is available by running a quick estimation.

        Returns
        -------
        bool
            True if GPU processing is available, False otherwise.

        """
        n_gpus = dev.get_device_count()
        if n_gpus > 0:
            # Test a very simple example to see if CuPy is working
            X = np.array([[2, 1], [1, 3], [3, 1], [2, 4]])
            y = np.array([0, 1, 0, 1])
            model = MixedLogit()
            model.fit(X, y, varnames=["a", "b"], alts=["1", "2"], n_draws=500,
                      randvars={'a': 'n', 'b': 'n'}, maxiter=0, verbose=0)
            print("{} GPU device(s) available. xlogit will use "
                  "GPU processing".format(n_gpus))
            return True
        else:
            print("*** No GPU device found. Verify CuPy is properly installed")
            return False
